/*
// UNIT TESTS OF THE BOT
                               
                                                   brake  M1   brake  M2   brake  M3    LACT
INPUT: FULL BRAKE Command String                = ~+0001,+0000,+0001,+0000,+0001,+0000,+0000
       Release Brakes Command String            = ~+0000,+0000,+0000,+0000,+0000,+0000,+0000
       Full Motor Forward String                = ~+0000,+0100,+0000,+0100,+0000,+0000,+0000
       Full Motor Backward String               = ~+0000,-0100,+0000,-0100,+0000,+0000,+0000

Turn Right                                      = ~+0000,+0100,+0000,-0100,+0000,+0000,+0000
Turn Left                                       = ~+0000,-0100,+0000,+0100,+0000,+0000,+0000
       
       Left Motor Command String                = ~+0000,+0100,+0000,+0000,+0000,+0000,+0000
       Right Motor Command String               = ~+0000,+0000,+0000,+0100,+0000,+0000,+0000
       Flipper forwards MotorCommand String     = ~+0000,+0000,+0000,+0000,+0000,+0100,+0000
       Flipper backwards MotorCommand String    = ~+0000,+0000,+0000,+0000,+0000,-0100,+0000
       LACT forwards MotorCommand String        = ~+0000,+0000,+0000,+0000,+0000,+0000,+3200
       LACT backwards MotorCommand String       = ~+0000,+0000,+0000,+0000,+0000,+0000,-3200


OUTPUT: ir sensor array in cm : (fl, fr, l, r, bl, br) from front left to back right


--------- NEW TARGET Motor Values -----------
target_Motor1_Command: 0
target_Motor2_Command: 0
target_FlipperMotor_Command: 0
target_LACT_Command: 0
---------------------------------------
blip - 60 milliseconds elapsed --- ( 97, 68, 129, 204, 190, 83)

*/

//----- KeepAlive Variables -----
#define KEEP_ALIVE_TIMEOUT_1      10000 
#define KEEP_ALIVE_TIMEOUT_2      6 ////roughly every 2 seconds if we use 20 but lets try 2
unsigned long start, finished, elapsed;
int keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
int keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;

int SAFTEY_SHUTDOWN_FORWARD = 0;
int SAFTEY_SHUTDOWN_BACKWARD = 0;

//----- SerialComms Variables -----
char targetCommandString[] = {'0','0','0','0','0'};

int target_Motor1_Command = 0;
int target_Motor1_brake_command = 0;
int target_Motor2_Command = 0;
int target_Motor2_brake_command = 0;
int target_FlipperMotor_Command = 0;
int target_FlipperMotor_brake_command = 0;
int target_LACT_Command = 0;

//----- PIR Hookups -----
#define LONG_RANGE_PIR_ENABLE     5
#define LONG_RANGE_PIR_PIN        6


//----- IMU Hookups -----
#include "quaternionFilters.h"
#include "MPU9250.h"

MPU9250 myIMU;
#define IMU_PROCESSING_SPEED1 10000
#define IMU_PROCESSING_SPEED2 10        // 100 = .12 hz       50 should be = 0.06 hz (6-fps)
int IMU_Processing_Index1 = 0; //Clock is so fast we need 2 variables to choke the processing... im talkin MHz fast 
int IMU_Processing_Index2 = 0; //second choke variables that starts to turn nano seconds into mili seconds

//----- GPS Hookups -----
//#include <TinyGPS++.h>
//#include <SoftwareSerial.h>
//TinyGPSPlus gps;
//static const int RXPin = 4, TXPin = 3;

// The serial connection to the GPS device
//SoftwareSerial ss(RXPin, TXPin);




void setup() {
  // put your setup code here, to run once:
  Serial.begin( 250000 );

 
  Serial.println("iBot v1.0 by OrbitusRobotics");
  Serial.println("BEGIN HEAD STARTUP SEQUENCE");
  
  //----- PIR SETUP -----
  initialize_PIR();
  //----- IMU SETUP -----
  Wire.begin();     //Needed by IMU
  initialize_IMU();
  //------ GPS SETUP -----
  //initialize_GPS(); //setup on SoftwareSerial Lines
  //----------------------
  

  Serial.println("STARTUP SEQUENCE COMPLETE");
  Serial.println("iBot v1.0 ONLINE");
 
  start=millis();

}

void loop() {
  // put your main code here, to run repeatedly:
    keepAlive_SAFTEYSTOP();

    
    process_IMU();
    //process_PIR();
    //process_GPS();

    
    if(Serial.available() > 0) {
 
      Serial.print("Parsing Serial and resetting keepAlive\n");
      resetKeepAlive();
      
      byte b =  Serial.read();
      
      if (b == 126)
      {
        char charIn = 0;

        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_Motor1_brake_command = atoi(targetCommandString);
        Serial.print("target_Motor1_brake_command ");
        Serial.println(target_Motor1_brake_command );

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!

        //irReadOut();
        
        Serial.println("---------------------------------------");
        
      }
   }
}


//------ KEEP ALIVE FUNCTIONS ------

void keepAlive_SAFTEYSTOP()
{
  keepAlive_1--;
    if (keepAlive_1 <= 0)
    {
      //MAKE SURE TO SET MOTORS BACK TO ZERO AFTER THE TIMEOUT IN CASE OF FAILURE AT FULL SPEED
      keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
      keepAlive_2--;

      //Serial.print("blip - ");
      //irReadOut();
      //Serial.print(elapsed);
      //Serial.print(" milliseconds elapsed --- ");
      
      //Serial.println(keepAlive_2);
      //irReadOut_SAFTEYSTOP();
    }
    if (keepAlive_2 <= 0)
    {
      finished=millis();
      elapsed=finished-start;
      start=millis();
      Serial.print("KEEPALIVE HAS EXPIRED!!! SHUTTING DOWN CONNECTION NOW -- ");
      Serial.print(elapsed);
      Serial.println(" milliseconds");
      keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
      
      //setMotorCommand(0,+0000, 0,+0000 ,0,+0000, +0000);
    }
}

void resetKeepAlive()
{
  keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
  keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
}

//------ IMU Functions -----

void initialize_IMU()
{
  // Read the WHO_AM_I register, this is a good test of communication
  byte c = myIMU.readByte(MPU9250_ADDRESS, WHO_AM_I_MPU9250);
  Serial.print(F("MPU9250 I AM 0x"));
  Serial.print(c, HEX);
  Serial.print(F(" I should be 0x"));
  Serial.println(0x71, HEX);

  if (c == 0x71) // WHO_AM_I should always be 0x71
  {
    Serial.println(F("MPU9250 is online..."));
    myIMU.calibrateMPU9250(myIMU.gyroBias, myIMU.accelBias);
    myIMU.initMPU9250();
    myIMU.initAK8963(myIMU.factoryMagCalibration);
    myIMU.getAres();
    myIMU.getGres();
    myIMU.getMres();
    Serial.println("MPU9250 initialized for active data mode....");
  } // if (c == 0x71)
  else
  {
    Serial.print("Could not connect to MPU9250: 0x");
    Serial.println(c, HEX);

    // Communication failed, stop here
    Serial.println(F("Communication failed, abort!"));
    Serial.flush();
    abort();
  }
}


void process_IMU()
{
    // This is a hardware like bitwise speed choke to turn MHz into kilo hertz.... or nanoseconds into microseconds
    IMU_Processing_Index1++;
    if (IMU_Processing_Index1 > IMU_PROCESSING_SPEED1)
    {
       IMU_Processing_Index1 = 0;
       IMU_Processing_Index2++;
       process_PIR();
    }
    if (IMU_Processing_Index2 > IMU_PROCESSING_SPEED2)
    {
       IMU_Processing_Index2 = 0;
       Serial.println("IMU Pulse");
       calculateIMU();
    }
}


void calculateIMU()
{
  if (myIMU.readByte(MPU9250_ADDRESS, INT_STATUS) & 0x01)
  {
    myIMU.readAccelData(myIMU.accelCount);  // Read the x/y/z adc values
    myIMU.ax = (float)myIMU.accelCount[0] * myIMU.aRes; // - myIMU.accelBias[0];
    myIMU.ay = (float)myIMU.accelCount[1] * myIMU.aRes; // - myIMU.accelBias[1];
    myIMU.az = (float)myIMU.accelCount[2] * myIMU.aRes; // - myIMU.accelBias[2];

    myIMU.readGyroData(myIMU.gyroCount);  // Read the x/y/z adc values
    myIMU.gx = (float)myIMU.gyroCount[0] * myIMU.gRes;
    myIMU.gy = (float)myIMU.gyroCount[1] * myIMU.gRes;
    myIMU.gz = (float)myIMU.gyroCount[2] * myIMU.gRes;

    myIMU.readMagData(myIMU.magCount);  // Read the x/y/z adc values
    myIMU.mx = (float)myIMU.magCount[0] * myIMU.mRes
               * myIMU.factoryMagCalibration[0] - myIMU.magBias[0];
    myIMU.my = (float)myIMU.magCount[1] * myIMU.mRes
               * myIMU.factoryMagCalibration[1] - myIMU.magBias[1];
    myIMU.mz = (float)myIMU.magCount[2] * myIMU.mRes
               * myIMU.factoryMagCalibration[2] - myIMU.magBias[2];
  }

  myIMU.updateTime();

  MahonyQuaternionUpdate(myIMU.ax, myIMU.ay, myIMU.az, myIMU.gx * DEG_TO_RAD,
                         myIMU.gy * DEG_TO_RAD, myIMU.gz * DEG_TO_RAD, myIMU.my,
                         myIMU.mx, myIMU.mz, myIMU.deltat);

  myIMU.delt_t = millis() - myIMU.count;
  if (myIMU.delt_t > 100)
  {
    ///*
    Serial.print("ax = ");  Serial.print((int)1000 * myIMU.ax);
    Serial.print(" ay = "); Serial.print((int)1000 * myIMU.ay);
    Serial.print(" az = "); Serial.print((int)1000 * myIMU.az);
    Serial.println(" mg");

    Serial.print("gx = ");  Serial.print(myIMU.gx, 2);
    Serial.print(" gy = "); Serial.print(myIMU.gy, 2);
    Serial.print(" gz = "); Serial.print(myIMU.gz, 2);
    Serial.println(" deg/s");

    Serial.print("mx = ");  Serial.print((int)myIMU.mx);
    Serial.print(" my = "); Serial.print((int)myIMU.my);
    Serial.print(" mz = "); Serial.print((int)myIMU.mz);
    Serial.println(" mG");

    Serial.print("q0 = ");  Serial.print(*getQ());
    Serial.print(" qx = "); Serial.print(*(getQ() + 1));
    Serial.print(" qy = "); Serial.print(*(getQ() + 2));
    Serial.print(" qz = "); Serial.println(*(getQ() + 3));
    //*/
    myIMU.yaw   = atan2(2.0f * (*(getQ()+1) * *(getQ()+2) + *getQ()
                  * *(getQ()+3)), *getQ() * *getQ() + *(getQ()+1)
                  * *(getQ()+1) - *(getQ()+2) * *(getQ()+2) - *(getQ()+3)
                  * *(getQ()+3));
    myIMU.pitch = -asin(2.0f * (*(getQ()+1) * *(getQ()+3) - *getQ()
                  * *(getQ()+2)));
    myIMU.roll  = atan2(2.0f * (*getQ() * *(getQ()+1) + *(getQ()+2)
                  * *(getQ()+3)), *getQ() * *getQ() - *(getQ()+1)
                  * *(getQ()+1) - *(getQ()+2) * *(getQ()+2) + *(getQ()+3)
                  * *(getQ()+3));
    myIMU.pitch *= RAD_TO_DEG;
    myIMU.yaw   *= RAD_TO_DEG;
    myIMU.yaw  -= 3.75; //True North for Austin, TX
    myIMU.roll *= RAD_TO_DEG;
    ///*
    Serial.print("Yaw, Pitch, Roll: ");
    Serial.print(myIMU.yaw, 2);
    Serial.print(", ");
    Serial.print(myIMU.pitch, 2);
    Serial.print(", ");
    Serial.println(myIMU.roll, 2);
    //*/
    myIMU.tempCount = myIMU.readTempData();  // Read the adc values
    // Temperature in degrees Centigrade
    myIMU.temperature = ((float) myIMU.tempCount) / 333.87 + 21.0;
    // Print temperature in degrees Centigrade
    ///*
    Serial.print("Temperature is ");  Serial.print(myIMU.temperature, 1);
    Serial.println(" degrees C");
    
    Serial.print("rate = ");
    Serial.print((float)myIMU.sumCount / myIMU.sum, 2);
    Serial.println(" Hz");
    //*/
    myIMU.count = millis();
    myIMU.sumCount = 0;
    myIMU.sum = 0;
  }
}


//------ GPS Functions -----
/*

void initialize_GPS()
{
  ss.begin(9600);
}


void process_GPS()
{
  while (ss.available() > 0)
    if (gps.encode(ss.read()))
      displayInfo();
  
  if (millis() > 5000 && gps.charsProcessed() < 10)
  {
    Serial.println(F("No GPS detected: check wiring."));
  }
}


void displayInfo()
{
  Serial.print(F("Location: ")); 
  if (gps.location.isValid())
  {
    Serial.print(gps.location.lat(), 6);
    Serial.print(F(","));
    Serial.print(gps.location.lng(), 6);
  }
  else
  {
    Serial.print(F("INVALID"));
  }

  Serial.print(F("  Date/Time: "));
  if (gps.date.isValid())
  {
    Serial.print(gps.date.month());
    Serial.print(F("/"));
    Serial.print(gps.date.day());
    Serial.print(F("/"));
    Serial.print(gps.date.year());
  }
  else
  {
    Serial.print(F("INVALID"));
  }

  Serial.print(F(" "));
  if (gps.time.isValid())
  {
    if (gps.time.hour() < 10) Serial.print(F("0"));
    Serial.print(gps.time.hour());
    Serial.print(F(":"));
    if (gps.time.minute() < 10) Serial.print(F("0"));
    Serial.print(gps.time.minute());
    Serial.print(F(":"));
    if (gps.time.second() < 10) Serial.print(F("0"));
    Serial.print(gps.time.second());
    Serial.print(F("."));
    if (gps.time.centisecond() < 10) Serial.print(F("0"));
    Serial.print(gps.time.centisecond());
  }
  else
  {
    Serial.print(F("INVALID"));
  }

  Serial.println();
}

*/
//------ PIR Functions -----


void initialize_PIR()
{
  pinMode( LONG_RANGE_PIR_ENABLE, OUTPUT);

  digitalWrite( LONG_RANGE_PIR_ENABLE, HIGH); //LOW is off

  pinMode( LONG_RANGE_PIR_PIN, INPUT);   
}

void process_PIR()
{
  int pirState = digitalRead(LONG_RANGE_PIR_PIN);
  if (pirState == HIGH) {
    // Motion Detected
    Serial.println("MOTION DETECTED");
  } else {
    // No Motion Detected
    //Serial.println("NO MOTION DETECTED");
  }
  //delay (20);
}

//----------------------------------------------
 
